
--[[ ItemRack 1.1 9/2/05 Gello ]]--

--[[ SavedVariables ]]--

ItemRack_Users = {} -- per-user bar settings (position, orientation, scale, locked status, bar contents)

ItemRack_Settings = {			-- These settings are for all users:
	TooltipFollow = "OFF",		-- whether tooltip shows at pointer or default position
	CooldownNumbers = "OFF",	-- whether cooldowns show a number overlay
	Soulbound = "OFF",			-- whether menu limits items to soulbound only
	Bindings = "OFF",			-- whether key bindings is displayed
	MenuShift = "OFF",			-- whether Shift key needs held to open the menu
	Notify = "OFF",				-- (future use)
	ShowEmpty = "OFF",			-- whether to show empty slot in the menu
	FlipMenu = "OFF",			-- whether to display menu on the opposite side
	RightClick = "OFF",			-- whether right click sends to second slot
}

--[[ Misc Variables ]]--

BINDING_HEADER_ItemRack = "ItemRack"
ItemRack_Version = 1.1

-- defaults for ItemRack_Users
local ItemRackOpt_Defaults = {
	MainOrient = "HORIZONTAL",  -- direction of main bar, "HORIZONTAL" or "VERTICAL", menu orient always opposite
	MainScale = 1,				-- scale 0-1 of main bar and menu bar
	XPos = 400,					-- left position of main bar
	YPos = 350,					-- top position of main bar
	Locked = "OFF"				-- "ON" or "OFF", lock status of main bar (for all intents menu always locked)
}

local user = "default"  -- "Gello of Hyjal" or "Gello of Deathwing ", defined in PLAYER_ENTERING_WORLD

local ItemRack = {}

ItemRack.InvUpdateTime = 0 -- time (in seconds) since last inventory update event
ItemRack.InvUpdateLimit = .25 -- time (in seconds) of no updates before reaction happens

ItemRack.ScaleTime = 0 -- time (in seconds) since last scale update
ItemRack.ScaleLimit = .1 -- time (in seconds) between scale updates
ItemRack.FrameToScale = nil -- holds frame being scaled for ScaleUpdate

ItemRack.MenuTime = 0 -- time (in seconds) since menu was last reset
ItemRack.MenuLimit = .25 -- time (in seconds) before menu dismisses if we left it

ItemRack.BaggedItems = {} -- table containing items in bags to show in menu
ItemRack.NumberOfItems = 0 -- number of items in the menu
ItemRack.MaxItems = 30 -- maximum number of items that can display in the menu
ItemRack.InvOpen = nil -- which inventory slot has its menu open

ItemRack.TooltipTime = 0 -- time since last tooltip OnUpdate
ItemRack.TooltipLimit = 1 -- frequency (in seconds) of updating tooltip
ItemRack.TooltipOwner = nil -- (this) when tooltip created
ItemRack.TooltipType = nil -- "BAG" or "INVENTORY"
ItemRack.TooltipBag = nil -- bag number
ItemRack.TooltipSlot = nil -- bag or inventory slot number

ItemRack.CooldownTime = 0 -- time (in seconds) since cooldowns last updated
ItemRack.CooldownLimit = 1 -- time (in seconds) before cooldowns are updated
ItemRack.CurrentTime = GetTime()

ItemRack.ControlTime = 0 -- time (in seconds) since control window last checked
ItemRack.ControlLimit = .75 -- time (in seconds) after leaving before the control disappears while locked

ItemRack.MainDock = "" -- "TOPLEFT" "BOTTOMRIGHT" etc
ItemRack.MenuDock = ""

ItemRack.Queue = {} -- [slot]="ItemName" if an item in queue, ie [13]="Arcanite Dragonling", [slot]=nil for no item in queue

--[[ reference tables ]]--

ItemRack.OptInfo = {
	["ItemRack_Control_Rotate"] = { text="Rotate", tooltip="Change the orientation of the bar between vertical or horizontal." },
	["ItemRack_Control_Lock"] = { text="Lock", tooltip="Lock or unlock the window.  While locked, hold ALT and move the mouse over the bar to access controls." },
	["ItemRack_Control_Options"] = { text="Settings", tooltip="Change ItemRack settings." },
	["ItemRack_Opt_TooltipFollow"] = { text="Tooltips At Pointer", tooltip="Check this to make tooltips follow the pointer instead of displaying at the default location.", type="Check", info="TooltipFollow" },
	["ItemRack_Opt_CooldownNumbers"] = { text="Cooldown Numbers", tooltip="Check this to display the numerical time left in an item's cooldown.", type="Check", info="CooldownNumbers" },
	["ItemRack_Opt_Soulbound"] = { text="Hide Tradables", tooltip="Check this to only display soulbound, quest or conjured items in the menus.  Bind-On-Equip and tradable items will be ignored to prevent showing up as you loot through an instance.", type="Check", info="Soulbound" },
	["ItemRack_Opt_Bindings"] = { text="Key Bindings", tooltip="Check this to display any key bindings on the bar.", type="Check", info="Bindings" },
	["ItemRack_Opt_MenuShift"] = { text="Menu On Shift", tooltip="Check this to prevent the menu from appearing unless you're holding the Shift key.", type="Check", info="MenuShift" },
	["ItemRack_Opt_Close"] = { text="Close Options", tooltip="Exit options.  To close ItemRack completely, remove all items from the bar or /itemrack." },
	["ItemRack_InvFrame_Resize"] = { text="Resize", tooltip="Drag the 'grip' in the lower right corner to resize the window." },
	["ItemRack_Opt_ShowEmpty"] = { text="Allow Empty Slots", tooltip="Check this to add empty slots to the swap menu.  Items will be dropped in the left-most available bag slot, if one exists.  You cannot queue empty slots.  Trinkets will not show empty slots if TrinketMenu Clicks is enabled.", type="Check", info="ShowEmpty" },
	["ItemRack_Opt_FlipMenu"] = { text="Flip Menu", tooltip="The direction the menu grows is perpendicular to the bar.  Check this to make it grow from the opposite side that it automatically chooses.", type="Check", info="FlipMenu" },
	["ItemRack_Opt_RightClick"] = { text="TrinketMenu Clicks", tooltip="Check this to make left-clicking a trinket go to the top trinket slot (as displayed on your character screen), and right-click to go to the bottom trinket slot.  (For those used to TrinketMenu)", type="Check", info="RightClick" }
}

-- paperdoll_slot=frame name of the slot on the paperdoll frame (for alt+click purposes)
-- ["SlotName"]=1 for each allowable slot
-- swappable=1 for slots that can be swapped in combat
-- ignore_soulbound = ignore soulbound flag for this slot
ItemRack.Indexes = {
	[0] = { paperdoll_slot="CharacterAmmoSlot", keybind="Use Ammo Item", ignore_soulbound=1, swappable=1, ["Ammo"]=1 },
	[1] = { paperdoll_slot="CharacterHeadSlot", keybind="Use Head Item", ["Head"]=1 },
	[2] = { paperdoll_slot="CharacterNeckSlot", keybind="Use Neck Item", ["Neck"]=1 },
	[3] = { paperdoll_slot="CharacterShoulderSlot", keybind="Use Shoulder Item", ["Shoulder"]=1 },
	[4] = { paperdoll_slot="CharacterShirtSlot", keybind="Use Shirt Item", ignore_soulbound=1, ["Shirt"]=1 },
	[5] = { paperdoll_slot="CharacterChestSlot", keybind="Use Chest Item", ["Chest"]=1 },
	[6] = { paperdoll_slot="CharacterWaistSlot", keybind="Use Waist Item", ["Waist"]=1 },
	[7] = { paperdoll_slot="CharacterLegsSlot", keybind="Use Legs Item", ["Legs"]=1 },
	[8] = { paperdoll_slot="CharacterFeetSlot", keybind="Use Feet Item", ["Feet"]=1 },
	[9] = { paperdoll_slot="CharacterWristSlot", keybind="Use Wrist Item", ["Wrist"]=1 },
	[10]= { paperdoll_slot="CharacterHandsSlot", keybind="Use Hands Item", ["Hands"]=1 },
	[11]= { paperdoll_slot="CharacterFinger0Slot", keybind="Use Top Finger Item", ["Finger"]=1 },
	[12]= { paperdoll_slot="CharacterFinger1Slot", keybind="Use Bottom Finger Item", ["Finger"]=1 },
	[13]= { paperdoll_slot="CharacterTrinket0Slot", ignore_soulbound=1, keybind="Use Top Trinket Item", ["Trinket"]=1 },
	[14]= { paperdoll_slot="CharacterTrinket1Slot", ignore_soulbound=1, keybind="Use Bottom Trinket Item", ["Trinket"]=1 },
	[15]= { paperdoll_slot="CharacterBackSlot", keybind="Use Back Item", ["Back"]=1 },
	[16]= { paperdoll_slot="CharacterMainHandSlot", keybind="Use Main-Hand Item", swappable=1, ["Main Hand"]=1, ["Two-Hand"]=1, ["One-Hand"]=1 },
	[17]= { paperdoll_slot="CharacterSecondaryHandSlot", keybind="Use Off-Hand Item", swappable=1, ["Off Hand"]=1, ["Held In Off-hand"]=1, ["One-Hand"]=1 },
	[18]= { paperdoll_slot="CharacterRangedSlot", keybind="Use Range Item", swappable=1, ["Ranged"]=1 },
	[19]= { paperdoll_slot="CharacterTabardSlot", keybind="Use Tabard Item", ignore_soulbound=1, ["Tabard"]=1 }
}

-- after 1.7 patch, equippable slots are from GetItemInfo which is not localized, until then, slots are localized
-- unfortunately GetItemInfo won't return equippable slot for range or projectiles on test yet, so stay with old method
-- when GetItemInfo returns range/ammo types, search and uncomment all "--1.7uncomment"

--if not LoadAddOn then --1.7uncomment

	-- this table will be obsolete after 1.7 patch, used to scan tooltip to see where item can be equipped. Localized
	ItemRack.Equipable = {
		[AMMOSLOT]=1, [INVTYPE_HEAD]=1, [INVTYPE_NECK]=1, [INVTYPE_SHOULDER]=1, [INVTYPE_BODY]=1, [INVTYPE_CHEST]=1,
		[INVTYPE_WAIST]=1, [INVTYPE_FEET]=1, [INVTYPE_WRIST]=1, [INVTYPE_FINGER]=1, [INVTYPE_TRINKET]=1, [INVTYPE_HAND]=1,
		[INVTYPE_CLOAK]=1, [INVTYPE_WEAPONMAINHAND]=1, [INVTYPE_2HWEAPON]=1, [INVTYPE_WEAPON]=1, [INVTYPE_WEAPONOFFHAND]=1,
		[INVTYPE_HOLDABLE]=1, [INVTYPE_RANGED]=1, [INVTYPE_TABARD]=1, [INVTYPE_LEGS]=1, [ITEMRACK_INVTYPE_PROJECTILE]=1,
		[ITEMRACK_INVTYPE_WAND]=1, [ITEMRACK_INVTYPE_GUN]=1, [ITEMRACK_INVTYPE_BOW]=1, [ITEMRACK_INVTYPE_CROSSBOW]=1
	}

	ItemRack.Indexes[0][AMMOSLOT] = 1
	ItemRack.Indexes[0][ITEMRACK_INVTYPE_PROJECTILE] = 1
	ItemRack.Indexes[1][INVTYPE_HEAD] = 1
	ItemRack.Indexes[2][INVTYPE_NECK] = 1
	ItemRack.Indexes[3][INVTYPE_SHOULDER] = 1
	ItemRack.Indexes[4][INVTYPE_BODY] = 1
	ItemRack.Indexes[5][INVTYPE_CHEST] = 1
	ItemRack.Indexes[6][INVTYPE_WAIST] = 1
	ItemRack.Indexes[7][INVTYPE_LEGS] = 1
	ItemRack.Indexes[8][INVTYPE_FEET] = 1
	ItemRack.Indexes[9][INVTYPE_WRIST] = 1
	ItemRack.Indexes[10][INVTYPE_HAND] = 1
	ItemRack.Indexes[11][INVTYPE_FINGER] = 1
	ItemRack.Indexes[12][INVTYPE_FINGER] = 1
	ItemRack.Indexes[13][INVTYPE_TRINKET] = 1
	ItemRack.Indexes[14][INVTYPE_TRINKET] = 1
	ItemRack.Indexes[15][INVTYPE_CLOAK] = 1
	ItemRack.Indexes[16][INVTYPE_WEAPONMAINHAND] = 1
	ItemRack.Indexes[16][INVTYPE_2HWEAPON] = 1
	ItemRack.Indexes[16][INVTYPE_WEAPON] = 1
	ItemRack.Indexes[17][INVTYPE_WEAPONOFFHAND] = 1
	ItemRack.Indexes[17][INVTYPE_HOLDABLE] = 1
	ItemRack.Indexes[17][INVTYPE_WEAPON] = 1
	ItemRack.Indexes[18][INVTYPE_RANGED] = 1
	ItemRack.Indexes[18][ITEMRACK_INVTYPE_WAND] = 1
	ItemRack.Indexes[18][ITEMRACK_INVTYPE_GUN] = 1
	ItemRack.Indexes[18][ITEMRACK_INVTYPE_BOW] = 1
	ItemRack.Indexes[18][ITEMRACK_INVTYPE_CROSSBOW] = 1
	ItemRack.Indexes[19][INVTYPE_TABARD] = 1

--end --1.7uncomment

-- dock-dependant offset and directions: MainDock..MenuDock
-- x/yoff   = offset MenuFrame is positioned to InvFrame
-- x/ydir   = direction items are added to menu
-- x/ystart = starting offset when building a menu, relativePoint MenuDock
-- mx/y     = offset MenuFrame is positioned to contents of InvFrame
local dock_stats = { ["TOPRIGHTTOPLEFT"] =		 { xoff=-4, yoff=0,  xdir=1,  ydir=-1, xstart=8,   ystart=-8,  mx=3,  my=8 },
					 ["BOTTOMRIGHTBOTTOMLEFT"] = { xoff=-4, yoff=0,  xdir=1,  ydir=1,  xstart=8,   ystart=44,  mx=3,  my=-8 },
					 ["TOPLEFTTOPRIGHT"] =		 { xoff=4,  yoff=0,  xdir=-1, ydir=-1, xstart=-44, ystart=-8,  mx=-2, my=8 },
					 ["BOTTOMLEFTBOTTOMRIGHT"] = { xoff=4,  yoff=0,  xdir=-1, ydir=1,  xstart=-44, ystart=44,  mx=-2, my=-8 },
					 ["TOPRIGHTBOTTOMRIGHT"] =   { xoff=0,  yoff=-4, xdir=-1, ydir=1,  xstart=-44,  ystart=44, mx=8,  my=3 },
					 ["BOTTOMRIGHTTOPRIGHT"] =   { xoff=0,  yoff=4,	 xdir=-1, ydir=-1, xstart=-44,  ystart=-8, mx=8,  my=-3 },
					 ["TOPLEFTBOTTOMLEFT"] =	 { xoff=0,  yoff=-4, xdir=1,  ydir=1,  xstart=8,   ystart=44,  mx=-8, my=2 },
					 ["BOTTOMLEFTTOPLEFT"] =	 { xoff=0,  yoff=4,  xdir=1,  ydir=-1, xstart=8,   ystart=-8,  mx=-8, my=-2 } }

-- returns offset and direction depending on current docking. ie: dock_info("xoff")
local function dock_info(which)

	local anchor = ItemRack.MainDock..ItemRack.MenuDock

	if dock_stats[anchor] and which and dock_stats[anchor][which] then
		return dock_stats[anchor][which]
	else
		return 0
	end
end

-- places the menu against the invslot
local function dock_menu(invslot)

	local item = getglobal("ItemRackInv"..invslot)
	local noflip = ItemRack_Settings.FlipMenu=="OFF"

	if ItemRack_Users[user].MainOrient=="HORIZONTAL" then
		if item:GetTop()*item:GetScale() < UIParent:GetHeight()/2 then
			if item:GetRight()*item:GetScale() < UIParent:GetWidth()/2 then
				ItemRack.MainDock = noflip and "TOPLEFT" or "BOTTOMLEFT"
				ItemRack.MenuDock = noflip and "BOTTOMLEFT" or "TOPLEFT"
			else
				ItemRack.MainDock = noflip and "TOPRIGHT" or "BOTTOMRIGHT"
				ItemRack.MenuDock = noflip and "BOTTOMRIGHT" or "TOPRIGHT"
			end
		else
			if item:GetRight()*item:GetScale() < UIParent:GetWidth()/2 then
				ItemRack.MainDock = noflip and "BOTTOMLEFT" or "TOPLEFT"
				ItemRack.MenuDock = noflip and "TOPLEFT" or "BOTTOMLEFT"
			else
				ItemRack.MainDock = noflip and "BOTTOMRIGHT" or "TOPRIGHT"
				ItemRack.MenuDock = noflip and "TOPRIGHT" or "BOTTOMRIGHT"
			end
		end
	else
		if item:GetTop()*item:GetScale() < UIParent:GetHeight()/2 then
			if item:GetRight()*item:GetScale() < UIParent:GetWidth()/2 then
				ItemRack.MainDock = noflip and "BOTTOMRIGHT" or "BOTTOMLEFT"
				ItemRack.MenuDock = noflip and "BOTTOMLEFT" or "BOTTOMRIGHT"
			else
				ItemRack.MainDock = noflip and "BOTTOMLEFT" or "BOTTOMRIGHT"
				ItemRack.MenuDock = noflip and "BOTTOMRIGHT" or "BOTTOMLEFT"
			end
		else
			if item:GetRight()*item:GetScale() < UIParent:GetWidth()/2 then
				ItemRack.MainDock = noflip and "TOPRIGHT" or "TOPLEFT"
				ItemRack.MenuDock = noflip and "TOPLEFT" or "TOPRIGHT"
			else
				ItemRack.MainDock = noflip and "TOPLEFT" or "TOPRIGHT"
				ItemRack.MenuDock = noflip and "TOPRIGHT" or "TOPLEFT"
			end
		end
	end

	ItemRack_MenuFrame:ClearAllPoints()
	ItemRack_MenuFrame:SetPoint(ItemRack.MenuDock,"ItemRackInv"..invslot,ItemRack.MainDock,dock_info("mx"),dock_info("my"))
end

-- returns texture,name,equipslot of an item in a container or inventory
local function get_item_info(bag,slot)

	local link,item,texture,name,equipslot,soulbound

	if slot then
		link = GetContainerItemLink(bag,slot)
		texture = GetContainerItemInfo(bag,slot)
	else
		link = GetInventoryItemLink("player",bag)
		texture = GetInventoryItemTexture("player",bag)
	end

	if link then
		_,_,item = string.find(link,"item:(%d+)")
		_,_,name = string.find(link,"^.*%[(.*)%].*$")
		_,_,_,_,_,_,_,equipslot = GetItemInfo(item)

--		if not LoadAddOn or ItemRack_Users[user].Soulbound=="ON" then --1.7uncomment
			-- if we're on a non-1.7 client, or we need to determine soulbound info, scan tooltip
			local i,text
			if slot then
				ItemRack_Tooltip:SetBagItem(bag,slot)
			else
				ItemRack_Tooltip:SetInventoryItem("player",bag)
			end
			for i=2,5 do
				text = tostring(getglobal("ItemRack_TooltipTextLeft"..i):GetText())
				if ItemRack.Equipable[text] then
					equipslot = text
				end
				if text==ITEM_SOULBOUND or text==ITEM_BIND_QUEST or text==ITEM_CONJURED then
					soulbound=true
				end
			end
--		end --1.7uncomment

		return texture,name,equipslot,soulbound
	elseif not slot then
		_,texture = GetInventorySlotInfo(string.gsub(ItemRack.Indexes[bag].paperdoll_slot,"Character",""))

		if bag==0 then
			-- this is a projectile, and isn't really equipped.  Get its texture from bags
			ItemRack_Tooltip:SetInventoryItem("player",0)
			text = tostring(ItemRack_TooltipTextLeft1:GetText())
			local i,j
			for i=0,4 do
				for j=1,GetContainerNumSlots(i) do
					_,_,item=string.find(GetContainerItemLink(i,j) or "","^.*%[(.*)%].*$")
					if item==text then
						texture = GetContainerItemInfo(i,j)
						j,i = GetContainerNumSlots(i)+1,5
					end
				end
			end
		end
		
		return texture
	end

	return nil
end

local function cursor_empty()
	return not (CursorHasItem() or CursorHasMoney() or CursorHasSpell())
end

-- returns true if frame is completely on screen
local function onscreen(frame)

	local scale = frame:GetParent():GetScale()
	local left,right,top,bottom = (frame:GetLeft() or 0)*scale, (frame:GetRight() or 0)*scale, (frame:GetTop() or 0)*scale, (frame:GetBottom() or 0)*scale

	if left>0 and right<1024 and top<768 and bottom>0 then
		return true
	else
		return false
	end
end

-- anything that may cause icons to go offscreen should be preceded by a hide_cooldowns()
local function hide_cooldowns()

	local i

	for i=0,19 do
		getglobal("ItemRackInv"..i.."Cooldown"):Hide()
	end

	for i=1,30 do
		getglobal("ItemRackMenu"..i.."Cooldown"):Hide()
	end
end

-- updates cooldown spinners in the menu
local function update_menu_cooldowns()

	local i, start, duration, enable

	if ItemRack.InvOpen then
		for i=1,ItemRack.NumberOfItems do
			if onscreen(getglobal("ItemRackMenu"..i)) then
				if ItemRack.BaggedItems[i].bag then
					start, duration, enable = GetContainerItemCooldown(ItemRack.BaggedItems[i].bag,ItemRack.BaggedItems[i].slot)
					CooldownFrame_SetTimer(getglobal("ItemRackMenu"..i.."Cooldown"), start, duration, enable)
				end
			end
		end
	end
end

-- updates cooldown spinners in the main bar
local function update_inv_cooldowns()

	local i, start, duration, enable

	if table.getn(ItemRack_Users[user].Bar)>0 then
		for i=1,table.getn(ItemRack_Users[user].Bar) do
			if onscreen(getglobal("ItemRackInv"..ItemRack_Users[user].Bar[i].."Cooldown")) then
				start, duration, enable = GetInventoryItemCooldown("player",ItemRack_Users[user].Bar[i])
				CooldownFrame_SetTimer(getglobal("ItemRackInv"..ItemRack_Users[user].Bar[i].."Cooldown"), start, duration, enable)
			end
		end
	end

	update_menu_cooldowns()
end

-- call this when window has changed and cooldowns need redrawn
local function cooldowns_need_updating()

	ItemRack.CooldownTime = ItemRack.CooldownLimit - 0.25 -- (in 1/4 second perform update)
	ItemRack.CooldownsNeedUpdating = true
	ItemRack_CooldownUpdate:Show()
end

local function populate_baggeditems(idx,bag,slot,name,texture)

	if not ItemRack.BaggedItems[idx] then
		ItemRack.BaggedItems[idx] = {}
	end
	ItemRack.BaggedItems[idx].bag = bag
	ItemRack.BaggedItems[idx].slot = slot
	ItemRack.BaggedItems[idx].name = name
	ItemRack.BaggedItems[idx].texture = texture
end

-- builds a menu outward from invslot (0-19)
local function build_menu(invslot)

	local idx,i,j,k,texture,name,equipslot,soulbound = 1

	dock_menu(invslot)

	for i=1,table.getn(ItemRack_Users[user].Bar) do
		if invslot~=ItemRack_Users[user].Bar[i] then
			getglobal("ItemRackInv"..ItemRack_Users[user].Bar[i]):UnlockHighlight()
		else
			getglobal("ItemRackInv"..ItemRack_Users[user].Bar[i]):LockHighlight()
		end
	end

	-- go through bags and gather items into .BaggedItems
	for i=0,4 do
		for j=1,GetContainerNumSlots(i) do
			texture,name,equipslot,soulbound = get_item_info(i,j)
			soulbound = soulbound or ItemRack.Indexes[invslot].ignore_soulbound -- pretend item soulbound if flagged to ignore_soulbound
			if (equipslot and ItemRack.Indexes[invslot][equipslot]) and (soulbound or ItemRack_Settings.Soulbound=="OFF") then
				populate_baggeditems(idx,i,j,name,texture)
				idx = idx + 1
			end
		end
	end

	if ItemRack_Settings.ShowEmpty=="ON" and GetInventoryItemLink("player",invslot) and not (ItemRack_Settings.RightClick=="ON" and (invslot==13 or invslot==14)) then
		-- add an empty slot to the menu
		_,_,i = string.find(ItemRack.Indexes[invslot].keybind,"Use (.+) Item")
		_,j = GetInventorySlotInfo(string.gsub(ItemRack.Indexes[invslot].paperdoll_slot,"Character",""))
		populate_baggeditems(idx,nil,nil,i or "",j)
		idx = idx + 1
	end

	ItemRack.NumberOfItems = math.min(idx-1,ItemRack.MaxItems)

	if ItemRack.NumberOfItems<1 then
		-- user has no bagged items for this type
		ItemRack_MenuFrame:Hide()
	else
		-- display items outward from docking point
		local col,row,xpos,ypos = 0,0,dock_info("xstart"),dock_info("ystart")
		local max_cols = 1

		if ItemRack.NumberOfItems>24 then
			max_cols = 5
		elseif ItemRack.NumberOfItems>18 then
			max_cols = 4
		elseif ItemRack.NumberOfItems>12 then
			max_cols = 3
		elseif ItemRack.NumberOfItems>4 then
			max_cols = 2
		end

		for i=1,ItemRack.NumberOfItems do
			local item = getglobal("ItemRackMenu"..i)
			getglobal("ItemRackMenu"..i.."Icon"):SetTexture(ItemRack.BaggedItems[i].texture)
			item:SetPoint("TOPLEFT","ItemRack_MenuFrame",ItemRack.MenuDock,xpos,ypos)

			if ItemRack_Users[user].MainOrient=="HORIZONTAL" then
				xpos = xpos + dock_info("xdir")*40
				col = col + 1
				if col==max_cols then
					xpos = dock_info("xstart")
					col = 0
					ypos = ypos + dock_info("ydir")*40
					row = row + 1
				end
				item:Show()
			else
				ypos = ypos + dock_info("ydir")*40
				col = col + 1
				if col==max_cols then
					ypos = dock_info("ystart")
					col = 0
					xpos = xpos + dock_info("xdir")*40
					row = row + 1
				end
				item:Show()
			end
		end
		for i=(ItemRack.NumberOfItems+1),ItemRack.MaxItems do
			getglobal("ItemRackMenu"..i):Hide()
		end
		if col==0 then
			row = row-1
		end

		if ItemRack_Users[user].MainOrient=="HORIZONTAL" then
			ItemRack_MenuFrame:SetWidth(12+(max_cols*40))
			ItemRack_MenuFrame:SetHeight(12+((row+1)*40))
		else
			ItemRack_MenuFrame:SetWidth(12+((row+1)*40))
			ItemRack_MenuFrame:SetHeight(12+(max_cols*40))
		end

		ItemRack.InvOpen = invslot
		ItemRack_MenuFrame:Show()
		update_menu_cooldowns()
		ItemRack.CooldownTime = ItemRack.CooldownLimit -- update cooldowns immediately

	end

end

-- for use with main/menu frames with UIParent parent when relocated by the mod, to register for layout-cache.txt
local function really_setpoint(frame,point,relativeTo,relativePoint,xoff,yoff)
	frame:SetPoint(point,relativeTo,relativePoint,xoff,yoff)
	frame:StartMoving()
	frame:StopMovingOrSizing()
	ItemRack_Users[user].XPos = xoff
	ItemRack_Users[user].YPos = yoff
end

-- draws the inventory bar
local function draw_inv()

	local xdir,ydir,corner,xadd,yadd,cx,cy,i,item,texture = 4,0,"TOPRIGHT",40,0,56,56

	if ItemRack_Users[user].MainOrient=="VERTICAL" then
		xdir,ydir,corner,xadd,yadd = 0,-4,"BOTTOMLEFT",0,40
	end

	for i=0,19 do
		getglobal("ItemRackInv"..i):Hide()
	end

	if table.getn(ItemRack_Users[user].Bar)>0 then
		item = getglobal("ItemRackInv"..ItemRack_Users[user].Bar[1])
		item:ClearAllPoints()
		item:SetPoint("TOPLEFT","ItemRack_InvFrame","TOPLEFT",10,-10)
		getglobal("ItemRackInv"..ItemRack_Users[user].Bar[1].."Icon"):SetTexture(get_item_info(ItemRack_Users[user].Bar[1]))
		item:Show()
		for i=2,table.getn(ItemRack_Users[user].Bar) do
			item = getglobal("ItemRackInv"..ItemRack_Users[user].Bar[i])
			item:ClearAllPoints()
			item:SetPoint("TOPLEFT","ItemRackInv"..ItemRack_Users[user].Bar[i-1],corner,xdir,ydir)
			getglobal("ItemRackInv"..ItemRack_Users[user].Bar[i].."Icon"):SetTexture(get_item_info(ItemRack_Users[user].Bar[i]))
			item:Show()
			cx = cx + xadd
			cy = cy + yadd
		end
		ItemRack_InvFrame:SetWidth(cx)
		ItemRack_InvFrame:SetHeight(cy)
		ItemRack_InvFrame:Show()
		cooldowns_need_updating()
		ItemRack.CooldownTime = ItemRack.CooldownLimit
	else
		ItemRack_InvFrame:Hide()
	end
end

local function unlocked()
	return ItemRack_Users[user].Locked~="ON"
end

-- sets window lock "ON" or "OFF"
local function set_lock(arg1)

	ItemRack_Users[user].Locked = arg1

	if arg1=="ON" then
		ItemRack_InvFrame:SetBackdropColor(0,0,0,0)
		ItemRack_InvFrame:SetBackdropBorderColor(0,0,0,0)
		ItemRack_InvFrame_Resize:Hide()
		ItemRack_Control_Rotate:SetAlpha(.4)
		ItemRack_Control_Rotate:Disable()
	else
		ItemRack_InvFrame:SetBackdropColor(1,1,1,1)
		ItemRack_InvFrame:SetBackdropBorderColor(1,1,1,1)
		ItemRack_InvFrame_Resize:Show()
		ItemRack_Control_Rotate:SetAlpha(1)
		ItemRack_Control_Rotate:Enable()
		ItemRack_ControlFrame:Show()
	end
end

-- swap item named 'name' into slot 'which' (0-19)
local function swap_by_name(which,name)

	local i,item,foundbag,foundslot

	if ItemRack.Queue[which] and cursor_empty() then
		for i=0,4 do
			for j=1,GetContainerNumSlots(i) do
				_,_,item = string.find(GetContainerItemLink(i,j) or "","^.*%[(.*)%].*$")
				if item==name then
					foundbag,foundslot = i,j
					j,i = GetContainerNumSlots(i)+1,5
				end
			end
		end
		if foundbag then
			PickupInventoryItem(which)
			PickupContainerItem(foundbag,foundslot)
			ItemRack.Swapping = true
		end
	end
end

-- called at startup, UPDATE_BINDINGS and option change to show/hide key bindings
local function update_keybindings()

	local i,modifier,key,text

	for i=0,19 do
		if ItemRack.Indexes[i].keybind and ItemRack_Settings.Bindings=="ON" then
			text = GetBindingKey(ItemRack.Indexes[i].keybind)
			_,_,modifier,key = string.find(text or "","(.).+(-.)")
			if modifier and key then
				text = modifier..key
			end
			getglobal("ItemRackInv"..i.."HotKey"):SetText(text)
		else
			getglobal("ItemRackInv"..i.."HotKey"):SetText("")
		end
	end
end

local function move_control()

	local item = ItemRack_InvFrame

	ItemRack_Control_Rotate:ClearAllPoints()
	ItemRack_Control_Lock:ClearAllPoints()
	ItemRack_Control_Options:ClearAllPoints()

	local midpoint = (table.getn(ItemRack_Users[user].Bar)*36 + 16)/2

	if ItemRack_Users[user].MainOrient=="HORIZONTAL" then
		if (ItemRack_Users[user].XPos+midpoint)*ItemRack_Users[user].MainScale < 512 then
			ItemRack_Control_Rotate:SetPoint("TOPLEFT","ItemRack_InvFrame","TOPRIGHT",-2,-3)
		else
			ItemRack_Control_Rotate:SetPoint("TOPRIGHT","ItemRack_InvFrame","TOPLEFT",2,-3)
		end
		ItemRack_Control_Lock:SetPoint("TOPLEFT","ItemRack_Control_Rotate","BOTTOMLEFT")
		ItemRack_Control_Options:SetPoint("TOPLEFT","ItemRack_Control_Lock","BOTTOMLEFT")
	else
		if (ItemRack_Users[user].YPos-midpoint)*ItemRack_Users[user].MainScale < 384 then
			ItemRack_Control_Rotate:SetPoint("BOTTOMLEFT","ItemRack_InvFrame","TOPLEFT",3,-2)
		else
			ItemRack_Control_Rotate:SetPoint("TOPLEFT","ItemRack_InvFrame","BOTTOMLEFT",3,2)
		end
		ItemRack_Control_Lock:SetPoint("TOPLEFT","ItemRack_Control_Rotate","TOPRIGHT")
		ItemRack_Control_Options:SetPoint("TOPLEFT","ItemRack_Control_Lock","TOPRIGHT")
	end
end

-- creates a user if one doesn't exist and initializes bar placement, scale, settings, etc from user settings
local function initialize()

	local i,item

	if not ItemRack_Users[user] then
		ItemRack_Users[user] = {} -- create new per-user setting
		ItemRack_Users[user].Inv = {}
		ItemRack_Users[user].Bar = {}
		for i in ItemRackOpt_Defaults do
			ItemRack_Users[user][i] = ItemRackOpt_Defaults[i]
		end
	end

	ItemRack_Settings.ShowEmpty = ItemRack_Settings.ShowEmpty or "OFF" -- 1.1
	ItemRack_Settings.FlipMenu = ItemRack_Settings.FlipMenu or "OFF" -- 1.1
	ItemRack_Settings.RightClick = ItemRack_Settings.RightClick or "OFF" -- 1.1

	ItemRack_InvFrame:SetScale(ItemRack_Users[user].MainScale or 1)
	ItemRack_MenuFrame:SetScale(ItemRack_Users[user].MainScale or 1)
	ItemRack_InvFrame:ClearAllPoints()
	really_setpoint(ItemRack_InvFrame,"TOPLEFT","UIParent","BOTTOMLEFT",ItemRack_Users[user].XPos,ItemRack_Users[user].YPos)
	draw_inv()
	set_lock(ItemRack_Users[user].Locked)
	update_keybindings()
	move_control()

	ItemRack_Opt_Title:SetText("ItemRack "..ItemRack_Version)

	for i in ItemRack.OptInfo do
		if ItemRack.OptInfo[i].type=="Check" then
			item = getglobal(i.."Text")
			item:SetText(ItemRack.OptInfo[i].text)
			item:SetTextColor(1,1,1)
			if ItemRack_Settings[ItemRack.OptInfo[i].info]=="ON" then
				getglobal(i):SetChecked(1)
			else
				getglobal(i):SetChecked(0)
			end
		end
	end

end

--[[ Frame functions ]]--

function ItemRack_OnLoad()

	oldItemRack_PaperDollItemSlotButton_OnClick = PaperDollItemSlotButton_OnClick
	PaperDollItemSlotButton_OnClick = newItemRack_PaperDollItemSlotButton_OnClick

	this:RegisterEvent("PLAYER_ENTERING_WORLD")
end

function ItemRack_OnEvent(event)

	if event=="UNIT_INVENTORY_CHANGED" then
		if arg1=="player" then
			ItemRack.InvUpdateTime=0
			ItemRack_InvUpdate:Show()
		end
	elseif (event=="PLAYER_REGEN_ENABLED" or event=="PLAYER_UNGHOST" or event=="PLAYER_ALIVE") and not UnitIsDeadOrGhost("player") then
		local i

		for i=0,19 do
			if ItemRack.Queue[i] then
				swap_by_name(i,ItemRack.Queue[i])
				ItemRack.Queue[i] = nil
				getglobal("ItemRackInv"..i.."Queue"):Hide()
			end
		end

	elseif event=="UPDATE_BINDINGS" then
		update_keybindings()

	elseif event=="PLAYER_ENTERING_WORLD" then
		this:UnregisterEvent("PLAYER_ENTERING_WORLD")

		user = UnitName("player").." of "..GetCVar("realmName")

		table.insert(UISpecialFrames,"ItemRack_OptFrame")

		initialize()

		SlashCmdList["ItemRackCOMMAND"] = ItemRack_SlashHandler
		SLASH_ItemRackCOMMAND1 = "/itemrack";

		this:RegisterEvent("PLAYER_REGEN_ENABLED")
		this:RegisterEvent("PLAYER_UNGHOST") -- called when alive after releasing
		this:RegisterEvent("PLAYER_ALIVE") -- called when alive before releasing
		this:RegisterEvent("UNIT_INVENTORY_CHANGED")
		this:RegisterEvent("UPDATE_BINDINGS")
	end
end

function ItemRack_SlashHandler(arg1)

	if not string.find(arg1,".+") then
		ItemRack_Toggle()
	elseif string.lower(arg1)=="reset" then
		ItemRack_Reset()
	elseif string.lower(arg1)=="lock" then
		set_lock("ON")
	elseif string.lower(arg1)=="unlock" then
		set_lock("OFF")
	else
		DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00ItemRack useage:")
		DEFAULT_CHAT_FRAME:AddMessage("/itemrack : toggle the window")
		DEFAULT_CHAT_FRAME:AddMessage("/itemrack reset : reset window")
		DEFAULT_CHAT_FRAME:AddMessage("/itemrack lock or unlock : toggles window lock")
		DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00Alt+click item slots in the character window to add/remove items.\nWhile locked, hold ALT over the window to access control buttons.")
	end
end

--[[ Main bar add/remove functions ]]--

function ItemRack_Reset()

	local i
	-- restore user settings to default (but keep bar contents)
	for i in ItemRackOpt_Defaults do
		ItemRack_Users[user][i] = ItemRackOpt_Defaults[i]
	end
	initialize()
end

local function remove_inv(id)

	local i

	getglobal("ItemRackInv"..id):Hide()
	ItemRack_Users[user].Inv[id] = nil
	for i=1,table.getn(ItemRack_Users[user].Bar) do
		if ItemRack_Users[user].Bar[i]==id then
			table.remove(ItemRack_Users[user].Bar,i)
		end
	end
	draw_inv()
	move_control()
end

-- Inv slots are added by ALT+clicking the paper doll
function newItemRack_PaperDollItemSlotButton_OnClick(button, ignoreShift)

	if IsAltKeyDown() then
		local item = string.gsub(tostring(this:GetName()),"Character","")
		local id,texture = GetInventorySlotInfo(item)

		if ItemRack_Users[user].Inv[id] then
			remove_inv(id)
		else
			ItemRack_Users[user].Inv[id] = 1
			getglobal("ItemRackInv"..id.."Icon"):SetTexture(texture)
			table.insert(ItemRack_Users[user].Bar,id)
			draw_inv()
			move_control()
		end
	else
		oldItemRack_PaperDollItemSlotButton_OnClick(button, ignoreShift)
	end
	
end

--[[ Inv Movement ]]--

function ItemRack_InvFrame_OnMouseDown(arg1)

	if arg1=="LeftButton" and ItemRack_Users[user].Locked=="OFF" then
		hide_cooldowns()
		ItemRack_OptFrame:Hide()
		this:StartMoving()
	end
end

function ItemRack_InvFrame_OnMouseUp(arg1)

	if arg1=="LeftButton" then
		this:StopMovingOrSizing()
		ItemRack_Users[user].XPos = ItemRack_InvFrame:GetLeft()
		ItemRack_Users[user].YPos = ItemRack_InvFrame:GetTop()
		cooldowns_need_updating()
		move_control()
	end
end

function ItemRack_Toggle()
	if ItemRack_InvFrame:IsVisible() then
		ItemRack_InvFrame:Hide()
	else
		ItemRack_InvFrame:Show()
	end
end

--[[ Inventory changes ]]--

function ItemRack_InvUpdate_OnUpdate(arg1)

	ItemRack.InvUpdateTime = ItemRack.InvUpdateTime + arg1
	if ItemRack.InvUpdateTime > ItemRack.InvUpdateLimit then
		ItemRack.InvUpdateTime = 0
		ItemRack_InvUpdate:Hide()
		ItemRack.Swapping = false
		draw_inv()
		if table.getn(ItemRack_Users[user].Bar)>0 then
			for i=1,table.getn(ItemRack_Users[user].Bar) do
				getglobal("ItemRackInv"..ItemRack_Users[user].Bar[i]):SetChecked(0)
				getglobal("ItemRackInv"..ItemRack_Users[user].Bar[i].."Icon"):SetVertexColor(1,1,1)
			end
		end
	end
end

--[[ Scaling ]]--

function ItemRack_StartScaling(arg1)
	if arg1=="LeftButton" and unlocked() then
		this:LockHighlight()
		hide_cooldowns()
		ItemRack.FrameToScale = this:GetParent()
		ItemRack.ScalingWidth = this:GetParent():GetWidth()
		ItemRack_MenuFrame:Hide()
		ItemRack_ScaleUpdate:Show()
	end
end

function ItemRack_StopScaling(arg1)
	if arg1=="LeftButton" then
		ItemRack_ScaleUpdate:Hide()
		ItemRack.FrameToScale = nil
		cooldowns_need_updating()
		this:UnlockHighlight()
		if this:GetParent():GetName() == "ItemRack_InvFrame" then
			ItemRack_Users[user].MainScale = ItemRack_InvFrame:GetScale()
		end
	end
end

function ItemRack_ScaleFrame(scale)
	local frame = ItemRack.FrameToScale
	local oldscale = frame:GetScale() or 1
	local framex = (frame:GetLeft() or ItemRack_Users[user].XPos)* oldscale
	local framey = (frame:GetTop() or ItemRack_Users[user].YPos)* oldscale

	frame:SetScale(scale)
	really_setpoint(ItemRack_InvFrame,"TOPLEFT","UIParent","BOTTOMLEFT",framex/scale,framey/scale)
end

function ItemRack_ScaleUpdate_OnUpdate(arg1)

	ItemRack.ScaleTime = ItemRack.ScaleTime + arg1
	if ItemRack.ScaleTime > ItemRack.ScaleLimit then
		ItemRack.ScaleTime = 0

		local frame = ItemRack.FrameToScale
		local oldscale = frame:GetScale()
		local framex, framey, cursorx, cursory = frame:GetLeft()*oldscale, frame:GetTop()*oldscale, GetCursorPosition()

		if (cursorx-framex)>32 then
			local newscale = (cursorx-framex)/ItemRack.ScalingWidth
			ItemRack_ScaleFrame(newscale)
		end

	end
end

--[[ Clicks ]]--

-- uses inventory slot v1 (0-19) can be called from key binding and slot doesn't need to be on the bar
function ItemRack_UseItem(v1)

	if v1 and not MerchantFrame:IsVisible() then
		if ItemRack_Users[user].Inv[v1] then
			getglobal("ItemRackInv"..v1):SetChecked(1)
		end
		UseInventoryItem(v1)
		ItemRack.InvUpdateTime = -1.25
		ItemRack_InvUpdate:Show()
	end

end

function ItemRack_Inv_OnClick(arg1)

	local id = this:GetID()

	this:SetChecked(0)

	if arg1=="LeftButton" and IsAltKeyDown() then
		-- if Alt is down, remove the item from the bar
		if ItemRack_Users[user].Inv[id] then
			remove_inv(id)
			ItemRack_MenuFrame:Hide()
		end
	elseif arg1=="LeftButton" and IsShiftKeyDown() and ChatFrameEditBox:IsVisible() then
		-- if Shift is down, link the item to chat
		ChatFrameEditBox:Insert(GetInventoryItemLink("player",id))
	else
		-- otherwise use the item
		ItemRack_UseItem(id)
	end
end

--[[ Menu ]]--

function ItemRack_Inv_OnEnter()
	if not ItemRack_ScaleUpdate:IsVisible() then
		if IsShiftKeyDown() or ItemRack_Settings.MenuShift=="OFF" then
			ItemRack.MenuTime = 0
			build_menu(this:GetID())
		end
		if IsAltKeyDown() then
			ItemRack_ControlFrame:Show()
		end
	end
end

function ItemRack_MenuFrame_OnUpdate(arg1)

	ItemRack.MenuTime = ItemRack.MenuTime + arg1
	if ItemRack.MenuTime > ItemRack.MenuLimit then
		ItemRack.MenuTime = 0
		if not MouseIsOver(ItemRack_InvFrame) and not MouseIsOver(ItemRack_MenuFrame) then
			ItemRack_MenuFrame:Hide()
		end
	end
end

function ItemRack_Menu_OnClick(arg1)

	local id,i,unqueue = this:GetID()

	this:SetChecked(0)

	if ItemRack_Settings.RightClick=="ON" and arg1=="LeftButton" and ItemRack.InvOpen==14 then
		ItemRack.InvOpen = 13
	elseif ItemRack_Settings.RightClick=="ON" and arg1=="RightButton" and ItemRack.InvOpen==13 then
		ItemRack.InvOpen = 14
	end

	if arg1=="LeftButton" and IsShiftKeyDown() and ChatFrameEditBox:IsVisible() then
		ChatFrameEditBox:Insert(GetContainerItemLink(ItemRack.BaggedItems[id].bag,ItemRack.BaggedItems[id].slot))

	elseif (UnitAffectingCombat("player") or UnitIsDeadOrGhost("player")) and not ItemRack.Indexes[ItemRack.InvOpen].swappable then

		if not ItemRack.BaggedItems[id].bag then
			-- user trying to queue an empty slot
			UIErrorsFrame:AddMessage("You can't queue an empty slot, sorry.",1,.25,.25,1,UIERRORS_HOLDTIME)
		else
			for i=0,19 do
				if ItemRack.Queue[i]==ItemRack.BaggedItems[id].name then
					ItemRack.Queue[i] = nil
					getglobal("ItemRackInv"..i.."Queue"):Hide()
					if i==ItemRack.InvOpen then
						unqueue = true
					end
				end
			end
			if not unqueue then
				ItemRack.Queue[ItemRack.InvOpen] = ItemRack.BaggedItems[id].name
				getglobal("ItemRackInv"..ItemRack.InvOpen.."Queue"):SetTexture(ItemRack.BaggedItems[id].texture)
				getglobal("ItemRackInv"..ItemRack.InvOpen.."Queue"):Show()
			end
		end
		ItemRack_MenuFrame:Hide()

	elseif ItemRack.InvOpen and cursor_empty() then
		ItemRack.Swapping = true
		if ItemRack.BaggedItems[id].bag then
			PickupContainerItem(ItemRack.BaggedItems[id].bag,ItemRack.BaggedItems[id].slot)
			PickupInventoryItem(ItemRack.InvOpen)
		else
			local j,bag,slot
			-- swapping to an empty slot, go find one
			for i=4,0,-1 do
				for j=1,GetContainerNumSlots(i) do
					if not GetContainerItemLink(i,j) then
						bag=i
						slot=j
						j,i = GetContainerNumSlots(i)+1,-1
					end
				end
			end
			if not bag then
				UIErrorsFrame:AddMessage(ERR_INV_FULL,1,.1,.1,1,UIERRORS_HOLDTIME)
			else
				PickupInventoryItem(ItemRack.InvOpen)
				PickupContainerItem(bag,slot)
			end
		end
		getglobal("ItemRackInv"..ItemRack.InvOpen.."Icon"):SetVertexColor(0.4,0.4,0.4)
		ItemRack_MenuFrame:Hide()
		ItemRack.InvUpdateTime = -1.25
		ItemRack_InvUpdate:Show()
	end
end

function ItemRack_MenuFrame_OnHide()

	local i
	for i=0,19 do
		getglobal("ItemRackInv"..i):UnlockHighlight()
	end
	ItemRack.InvOpen = nil
end

--[[ Tooltips ]]--

function ItemRack_Inv_Tooltip()

	local id = this:GetID()

	if ItemRack_ScaleUpdate:IsVisible() then
		return
	end

	ItemRack.TooltipOwner = this
	ItemRack.TooltipType = "INVENTORY"
	ItemRack.TooltipSlot = id

	ItemRack.TooltipBag = ItemRack.Queue[id]

	ItemRack.TooltipTime = ItemRack.TooltipLimit -- immediate update
	ItemRack_TooltipUpdate:Show()
end

function ItemRack_Menu_Tooltip()

	local id = this:GetID()

	if ItemRack_ScaleUpdate:IsVisible() then
		return
	end

	if ItemRack.BaggedItems[id].bag then
		ItemRack.TooltipOwner = this
		ItemRack.TooltipType = "BAG"
		ItemRack.TooltipBag = ItemRack.BaggedItems[id].bag
		ItemRack.TooltipSlot = ItemRack.BaggedItems[id].slot

		ItemRack.TooltipTime = ItemRack.TooltipLimit -- immediate update
		ItemRack_TooltipUpdate:Show()
	end
end

function ItemRack_ClearTooltip()

	GameTooltip:Hide()
	ItemRack.TooltipType = nil
	ItemRack_TooltipUpdate:Hide()
	if not ItemRack.InvOpen then
		ItemRack_MenuFrame_OnHide()
	end
end

local function set_tooltip_anchor(owner)

	if ItemRack_Settings.TooltipFollow=="ON" then
		if owner:GetLeft()<400 then
			GameTooltip:SetOwner(owner,"ANCHOR_RIGHT")
		else
			GameTooltip:SetOwner(owner,"ANCHOR_LEFT")
		end
	else
		GameTooltip_SetDefaultAnchor(GameTooltip,owner)
	end
end

-- updates the tooltip created in the functions above
function ItemRack_TooltipUpdate_OnUpdate(arg1)

	ItemRack.TooltipTime = ItemRack.TooltipTime + arg1
	if ItemRack.TooltipTime > ItemRack.TooltipLimit then
		ItemRack.TooltipTime = 0

		if ItemRack.TooltipType then

			local cooldown

			set_tooltip_anchor(ItemRack.TooltipOwner)

			if ItemRack.TooltipType=="BAG" then
				GameTooltip:SetBagItem(ItemRack.TooltipBag,ItemRack.TooltipSlot)
				cooldown = GetContainerItemCooldown(ItemRack.TooltipBag,ItemRack.TooltipSlot)
			else
				GameTooltip:SetInventoryItem("player",ItemRack.TooltipSlot)
				cooldown = GetInventoryItemCooldown("player",ItemRack.TooltipSlot)
				if ItemRack.TooltipBag then
					GameTooltip:AddLine("Queued: "..ItemRack.TooltipBag)
				end
			end
			GameTooltip:Show()
			if cooldown==0 then
				-- stop updates if this item has no cooldown
				ItemRack.TooltipType = nil
				ItemRack_TooltipUpdate:Hide()
			end
		end

	end

end

--[[ Cooldowns ]]--

local function format_time(seconds)

	if seconds<60 then
		return math.floor(seconds+.5).." s"
	else
		if seconds < 3600 then 
			return math.ceil((seconds/60)).." m"
		else
			return math.ceil((seconds/3600)).." h"
		end
	end

end

local function write_cooldown(where,start,duration)

	local cooldown = duration - (ItemRack.CurrentTime - start)

	if start==0 then
		where:SetText("")
	elseif cooldown<3 and not where:GetText() then
		-- this is a global cooldown. don't display it. not accurate but at least not annoying
	else
		where:SetText(format_time(cooldown))
	end
end

function ItemRack_CooldownUpdate_OnUpdate(arg1)

	ItemRack.CooldownTime = ItemRack.CooldownTime + arg1
	if ItemRack.CooldownTime > ItemRack.CooldownLimit then
		ItemRack.CooldownTime = 0
		-- update spinners if needed
		if ItemRack.CooldownsNeedUpdating then
			ItemRack.CooldownsNeedUpdating = false
			if ItemRack_Settings.CooldownNumbers=="OFF" and LoadAddOn then
				ItemRack_CooldownUpdate:Hide() -- hide if in 1.7, scaling bug fixed
			end
			update_inv_cooldowns()
		end
		-- fix for scaling bug on alt-tab: see Rowne's Switchfix at http://www.curse-gaming.com/mod.php?addid=1052
		-- hopefully not needed in 1.7 --1.7uncomment
		if UIParent:GetScale()==ItemRack_ScaleBugFix:GetScale() and ItemRack_Users[user].MainScale then
			ItemRack_ScaleBugFix:SetScale(0.2)
			ItemRack_InvFrame:SetScale(ItemRack_Users[user].MainScale)
			ItemRack_MenuFrame:SetScale(ItemRack_Users[user].MainScale)
		end

		if ItemRack_Settings.CooldownNumbers=="ON" then

			local i,start,duration
			ItemRack.CurrentTime = GetTime()

			if ItemRack_InvFrame:IsVisible() then
				for i=1,table.getn(ItemRack_Users[user].Bar) do
					start, duration = GetInventoryItemCooldown("player",ItemRack_Users[user].Bar[i])
					write_cooldown(getglobal("ItemRackInv"..ItemRack_Users[user].Bar[i].."Time"),start,duration)
				end
			end

			if ItemRack_MenuFrame:IsVisible() and ItemRack.InvOpen then
				for i=1,ItemRack.NumberOfItems do
					if ItemRack.BaggedItems[i].bag then
						start, duration = GetContainerItemCooldown(ItemRack.BaggedItems[i].bag,ItemRack.BaggedItems[i].slot)
						write_cooldown(getglobal("ItemRackMenu"..i.."Time"),start,duration)
					end
				end
			end

		end

	end
end

function ItemRack_CooldownUpdate_OnHide()

	local i

	for i=0,19 do
		getglobal("ItemRackInv"..i.."Time"):SetText("")
	end
	for i=1,30 do
		getglobal("ItemRackMenu"..i.."Time"):SetText("")
	end
end

--[[ Options ]]--

function ItemRack_OnTooltip(v1,v2)

	set_tooltip_anchor(this)
	GameTooltip:AddLine(v1)
	GameTooltip:AddLine(v2,.8,.8,.8,1)
	GameTooltip:Show()
end

function ItemRack_Opt_OnEnter()

	local id = this:GetName()

	if ItemRack.OptInfo[id] and not ItemRack_ScaleUpdate:IsVisible() then
		ItemRack_OnTooltip(ItemRack.OptInfo[id].text,ItemRack.OptInfo[id].tooltip)
	end
end

function ItemRack_Control_OnClick()

	local id = this:GetName()

	if id=="ItemRack_Control_Rotate" and ItemRack_Users[user].Locked=="OFF" then
		-- rotate the window
		ItemRack_Users[user].MainOrient = ItemRack_Users[user].MainOrient=="HORIZONTAL" and "VERTICAL" or "HORIZONTAL"
		hide_cooldowns()
		ItemRack_MenuFrame:Hide()
		draw_inv()
		move_control()
	elseif id=="ItemRack_Control_Lock" then
		-- lock/unlock the window
		ItemRack_Users[user].Locked = ItemRack_Users[user].Locked=="ON" and "OFF" or "ON"
		set_lock(ItemRack_Users[user].Locked)
	elseif id=="ItemRack_Control_Options" then
		if ItemRack_OptFrame:IsVisible() then
			ItemRack_OptFrame:Hide()
		else
			local item,corner,other = ItemRack_InvFrame
			if item:GetLeft()*item:GetScale() < 512 then
				if item:GetTop()*item:GetScale() < 384 then
					corner = "BOTTOMLEFT"
					other = "BOTTOMRIGHT"
				else
					corner = "TOPLEFT"
					other = "TOPRIGHT"
				end
			else
				if item:GetTop()*item:GetScale() < 384 then
					corner = "BOTTOMRIGHT"
					other = "BOTTOMLEFT"
				else
					corner = "TOPRIGHT"
					other = "TOPLEFT"
				end
			end
			ItemRack_OptFrame:ClearAllPoints()
			ItemRack_OptFrame:SetPoint(corner,"ItemRack_Control_Options",other)
			ItemRack_OptFrame:Show()
		end
	elseif id=="ItemRack_Opt_Close" then
		ItemRack_OptFrame:Hide()
	end
end

-- runs while control buttons are up -- if locked and alt not pressed it dismisses the window
function ItemRack_ControlFrame_OnUpdate(arg1)

	ItemRack.ControlTime = ItemRack.ControlTime + arg1
	if ItemRack.ControlTime>ItemRack.ControlLimit and ItemRack_Users[user].Locked=="ON" then
		ItemRack.ControlTime = 0
		if not IsAltKeyDown() then
			ItemRack_ControlFrame:Hide()
		end
	end
end

function ItemRack_Opt_OnClick()

	local id = this:GetName()

	if ItemRack.OptInfo[id] then
		if this:GetChecked() then
			ItemRack_Settings[ItemRack.OptInfo[id].info] = "ON"
			PlaySound("igMainMenuOptionCheckBoxOn")
		else
			ItemRack_Settings[ItemRack.OptInfo[id].info] = "OFF"
			PlaySound("igMainMenuOptionCheckBoxOff")
		end

		if id=="ItemRack_Opt_Bindings" then
			update_keybindings()
		elseif id=="ItemRack_Opt_CooldownNumbers" then
			ItemRack_CooldownUpdate_OnHide()
			ItemRack.CooldownTime = ItemRack.CooldownLimit -- immediate update
		end
	end
end
